
/* Chrysalide - Outil d'analyse de fichiers binaires
 * arithm.c - représentation des opérations arithmétiques
 *
 * Copyright (C) 2010-2013 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "arithm.h"


#include "../expression-int.h"



/* Définition d'une opération arithmétique définie (instance) */
struct _GArithmExpression
{
    GDecExpression parent;                  /* A laisser en premier        */

    GDecExpression *op1;                    /* Premier opérande manipulé   */
    GDecExpression *op2;                    /* Second opérande manipulé    */

    ArithmOperationType type;               /* Opération à représenter     */

};


/* Définition d'une opération arithmétique définie (classe) */
struct _GArithmExpressionClass
{
    GDecExpressionClass parent;             /* A laisser en premier        */

};



/* Initialise la classe des opérations arithmétiques définies. */
static void g_arithm_expression_class_init(GArithmExpressionClass *);

/* Initialise une instance d'opération arithmétique définie. */
static void g_arithm_expression_init(GArithmExpression *);

/* Visite un ensemble hiérarchique d'instructions décompilées. */
static bool g_arithm_expression_visit(GArithmExpression *, dec_instr_visitor_cb, DecInstrVisitFlags, void *);

/* Remplace une instruction décompilée par une autre. */
static bool g_arithm_expression_replace(GArithmExpression *, GDecInstruction *, GDecInstruction *);

/* Imprime pour l'écran un version humaine d'une expression. */
static GBufferLine *g_arithm_expression_print(const GArithmExpression *, GCodeBuffer *, GBufferLine *, GLangOutput *);



/* Indique le type défini pour une opération arithmétique définie. */
G_DEFINE_TYPE(GArithmExpression, g_arithm_expression, G_TYPE_DEC_EXPRESSION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des opérations arithmétiques définies.  *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arithm_expression_class_init(GArithmExpressionClass *klass)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = instance à initialiser.                               *
*                                                                             *
*  Description : Initialise une instance d'opération arithmétique définie.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arithm_expression_init(GArithmExpression *expr)
{
    GDecInstruction *instr;                 /* Autre version de l'objet    */

    instr = G_DEC_INSTRUCTION(expr);

    instr->visit = (dec_instr_visit_fc)g_arithm_expression_visit;
    instr->replace = (dec_instr_replace_fc)g_arithm_expression_replace;
    instr->print = (dec_instr_print_fc)g_arithm_expression_print;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op1  = premier opérande à manipuler.                         *
*                type = type d'opération à mener ici.                         *
*                op2  = seconde opérande à manipuler.                         *
*                                                                             *
*  Description : Représente une opération arithmétique entre deux opérandes.  *
*                                                                             *
*  Retour      : Expression mise en place.                                    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GDecInstruction *g_arithm_expression_new(GDecExpression *op1, ArithmOperationType type, GDecExpression *op2)
{
    GArithmExpression *result;              /* Expression à retourner      */

    result = g_object_new(G_TYPE_ARITHM_EXPRESSION, NULL);

    result->op1 = op1;
    result->op2 = op2;

    result->type = type;

    return G_DEC_INSTRUCTION(result);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr     = première instruction à venir visiter.             *
*                callback = procédure à appeler à chaque instruction visitée. *
*                flags    = moments des appels à réaliser en retour.          *
*                data     = données quelconques associées au visiteur.        *
*                                                                             *
*  Description : Visite un ensemble hiérarchique d'instructions décompilées.  *
*                                                                             *
*  Retour      : true si le parcours a été jusqu'à son terme, false sinon.    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_arithm_expression_visit(GArithmExpression *expr, dec_instr_visitor_cb callback, DecInstrVisitFlags flags, void *data)
{
    bool result;                            /* Bilan à retourner           */

    result = _g_dec_instruction_visit(G_DEC_INSTRUCTION(expr->op1), G_DEC_INSTRUCTION(expr),
                                      callback, flags, data);

    if (result)
        result = _g_dec_instruction_visit(G_DEC_INSTRUCTION(expr->op2), G_DEC_INSTRUCTION(expr),
                                          callback, flags, data);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = première instruction à venir ausculter.               *
*                old  = instruction décompilée à venir remplacer.             *
*                new  = instruction décompilée à utiliser dorénavant.         *
*                                                                             *
*  Description : Remplace une instruction décompilée par une autre.           *
*                                                                             *
*  Retour      : true si un remplacement a été effectué, false sinon.         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_arithm_expression_replace(GArithmExpression *expr, GDecInstruction *old, GDecInstruction *new)
{
    bool result;                            /* Bilan à retourner           */

    if (expr->op1 == G_DEC_EXPRESSION(old))
    {
        g_object_unref(G_OBJECT(expr->op1));
        g_object_ref(G_OBJECT(new));
        expr->op1 = G_DEC_EXPRESSION(new);

        result = true;

    }
    else
        result = g_dec_instruction_replace(G_DEC_INSTRUCTION(expr->op1), old, new);

    if (expr->op2 == G_DEC_EXPRESSION(old))
    {
        g_object_unref(G_OBJECT(expr->op2));
        g_object_ref(G_OBJECT(new));
        expr->op2 = G_DEC_EXPRESSION(new);

        result = true;

    }
    else
        result |= g_dec_instruction_replace(G_DEC_INSTRUCTION(expr->op2), old, new);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr   = expression à transcrire en version humaine.         *
*                buffer = tampon où doit se réaliser l'insertion.             *
*                line   = ligne d'impression prête à emploi ou NULL.          *
*                output = langage de programmation de sortie.                 *
*                                                                             *
*  Description : Imprime pour l'écran un version humaine d'une expression.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GBufferLine *g_arithm_expression_print(const GArithmExpression *expr, GCodeBuffer *buffer, GBufferLine *line, GLangOutput *output)
{
    GBufferLine *result;                    /* Ligne à retourner           */
    const char *sign;                       /* Symbole de l'opération      */

    result = g_dec_instruction_print(G_DEC_INSTRUCTION(expr->op1),
                                     buffer, line, output);

    switch (expr->type)
    {
        case AOT_ADD:
            sign = " + ";
            break;
        case AOT_SUB:
            sign = " - ";
            break;
        case AOT_MUL:
            sign = " * ";
            break;
        case AOT_DIV:
            sign = " / ";
            break;
        case AOT_REM:
            sign = " % ";
            break;
        case AOT_AND:
            sign = " & ";
            break;
        case AOT_OR:
            sign = " | ";
            break;
        case AOT_XOR:
            sign = " ^ ";
            break;
        default:    /* AOT_COUNT */
            sign = " ? ";
            break;
    }

    g_buffer_line_append_text(result, BLC_ASSEMBLY_HEAD, sign, 3, RTT_SIGNS, NULL);

    result = g_dec_instruction_print(G_DEC_INSTRUCTION(expr->op2),
                                     buffer, result, output);

    return result;

}
